==> file4life-1.1.3-source.tgz: file4life-1.1.3/src/F4L/MockFSMonad.hs <==
{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}

module F4L.MockFSMonad where

import F4L.BufferMonad
import F4L.FSMonad
import F4L.Utils( uuidString )

import Control.DeepSeq
import Control.Lens
import Control.Monad.Error( throwError )
import Control.Monad.State.Strict
import qualified Data.ByteString.Lazy as BL
import Data.List( inits, stripPrefix )
import qualified Data.Map.Strict as M
import Data.Maybe( fromMaybe, mapMaybe )
import qualified Data.Set as S
import Data.Set( Set )
import System.Console.Haskeline.MonadException
import System.FilePath( takeDirectory, joinPath, splitDirectories )


subpaths :: FilePath -> [ FilePath ]
subpaths = map joinPath . drop 1 . inits . splitDirectories

{------------------------------ FS ------------------------------}

class MockROFSState mfs where
  mfs_files :: mfs -> [FilePath]
  mfs_get_contents :: FilePath -> mfs -> Maybe BL.ByteString
  mfs_dirs :: mfs -> [FilePath]
  mfs_is_deleted :: FilePath -> mfs -> Bool
  mfs_is_deleted _ _ = False


class MockROFSState mfs => MockFSState mfs where
  mfs_set_contents :: FilePath -> BL.ByteString -> mfs -> mfs
  mfs_delete :: FilePath -> mfs -> mfs
  mfs_mkdir :: FilePath -> mfs -> mfs
  mfs_now :: Lens' mfs Int


mfs_modifyContents
  :: MockFSState mfs   
  => FilePath -> ( BL.ByteString -> BL.ByteString ) -> mfs -> mfs
mfs_modifyContents p f fs = mfs_set_contents p ( f $ fromMaybe BL.empty $ mfs_get_contents p fs ) fs

mfs_mkdirAndParents :: MockFSState mfs => FilePath -> mfs -> mfs
mfs_mkdirAndParents f fs =   foldr mfs_mkdir fs $ subpaths f


{------------------------------ Overlay of 2 FSs ------------------------------}

-- | The first parameter overrides the second
data Overlay a b = Overlay a b

instance ( MockROFSState a, MockROFSState b ) => MockROFSState ( Overlay a b )
  where
    mfs_files ( Overlay a b ) = S.toList $ S.fromList $ mfs_files a ++ filter ( not . flip mfs_is_deleted a ) ( mfs_files b )
    mfs_get_contents f ( Overlay a b ) = mfs_get_contents f a `mplus` mfs_get_contents f b
    mfs_dirs ( Overlay a b ) = S.toList $ S.fromList $ mfs_dirs a ++ mfs_dirs b
    mfs_is_deleted f ( Overlay a _ ) = mfs_is_deleted f a


instance ( MockFSState a, MockROFSState b ) => MockFSState ( Overlay a b )
  where
    mfs_set_contents f s ( Overlay a b ) = Overlay ( mfs_set_contents f s a ) b
    mfs_delete f ( Overlay a b ) = Overlay ( mfs_delete f a ) b
    mfs_mkdir f ( Overlay a b ) = Overlay ( mfs_mkdir f a ) b
    mfs_now = lens f g . mfs_now
      where f ( Overlay a _ ) = a
            g ( Overlay _ b ) a = Overlay a b


{------------------------------ Virtual FS ------------------------------}

data VirtualFS = VirtualFS
                 FilePath -- root
                 Int -- number of files

instance MockROFSState VirtualFS where
  mfs_files ( VirtualFS root n ) = [ joinPath [ root, show i ] | i <- [ 1 .. n ] ]
  mfs_dirs ( VirtualFS root _ ) = subpaths root
  mfs_get_contents f ( VirtualFS root n ) =
    case splitAt ( length root ) f of
      (a, '/' : x @ (_ : _)) | a == root ->
        if all ( `elem` ['0' .. '9'] ) x && ( read x >= 1 ) && ( read x <= n )
           then Just $ BL.pack $ map ( toEnum . fromEnum ) $ "File " ++ show ( read x :: Int )
                else Nothing
      _ -> Nothing

{------------------------------ MemFS ------------------------------}

data MemFS = MemFS { _ms_files :: M.Map FilePath BL.ByteString
                   , _ms_deleted :: Set FilePath
                   , _ms_dirs :: Set FilePath
                   , _ms_now :: Int
                   }
           deriving Show

instance MockROFSState MemFS where
  mfs_files = M.keys . view ms_files
  mfs_get_contents f = view $ ms_files . to (M.lookup f)
  mfs_dirs = S.toList . view ms_dirs
  mfs_is_deleted f = view $ ms_deleted . to (S.member f)


instance MockFSState MemFS where
  mfs_set_contents f b =
    BL.length b `seq` over ms_files $ M.insert f b
  mfs_delete f = over ms_files ( M.delete f ) . over ms_deleted ( S.insert f )
  mfs_mkdir p = over ms_dirs $ S.insert p
  mfs_now = lens _ms_now $ \ s a -> s { _ms_now = a }


emptyMemFS = MemFS M.empty S.empty ( S.singleton "/" ) 0


ms_files :: Lens' MemFS ( M.Map FilePath BL.ByteString )
ms_files = lens _ms_files ( \ s a -> s { _ms_files = a } )
ms_dirs :: Lens' MemFS ( Set FilePath )
ms_dirs = lens _ms_dirs ( \ s a -> s { _ms_dirs = a } )
ms_deleted :: Lens' MemFS ( Set FilePath )
ms_deleted = lens _ms_deleted ( \ s a -> s { _ms_deleted = a } )


{------------------------------ Monad ------------------------------}

newtype MockFS fs a = MockFS { deMockFS :: StateT fs IOWithBuffer a }
                    deriving ( Monad, MonadIO, MonadState fs, BufferMonad, ParFSMonad )


runMockFS :: MockFS fs a -> StateT fs IO a
runMockFS a = StateT $ runIOWithBuffer . runStateT ( deMockFS a )


instance MonadException ( MockFS mfs ) where
  controlIO _ = error "controlIO for MockFS not implemented"


instance MockFSState mfs => DelFSMonad ( MockFS mfs ) where
  removeFile = modify . mfs_delete


instance MockFSState mfs => FSMonad ( MockFS mfs ) where
  outputMsg = liftIO . outputMsg

  exifProperties = const $ return []

  prompt = liftIO . prompt

  generateUuid = liftIO uuidString

  getNow = use mfs_now

  readFromFile f = do
    mf <- gets $ mfs_get_contents f
    case mf of
      Nothing -> throwError $ "File " ++ f ++ " not found"
      Just s -> return $ map (toEnum . fromEnum) $ BL.unpack s

  actOnLengthAndBLFromFile f action = do
    mf <- gets $ mfs_get_contents f
    case mf of
      Nothing -> throwError $ "File " ++ f ++ " not found"
      Just s -> lift $ action ( fromIntegral $ BL.length s ) s
      
  appendToFile f s = do
    pre <- gets $ fromMaybe BL.empty . mfs_get_contents f
    modify $ mfs_set_contents f $ BL.append pre $ BL.pack $ map (toEnum . fromEnum) s
    -- re-read to make behaviour strictier. TODO: check if performance hit
  
  appendBLsToFile f bls = do
    pre <- gets $ fromMaybe BL.empty . mfs_get_contents f
    modify $ mfs_set_contents f $ foldl BL.append pre bls
  
  canonicalizePath = return -- liftIO . canonicalizePath
  
{-
  copyFile a b = do
    s <- gets $ mfs_get_contents a
    case s of
      Nothing -> fail $ "No such file " ++ a
      Just r -> modify $ mfs_set_contents b r
-}

  doesDirectoryExist d = gets mfs_dirs >>= return . elem d
  
  doesFileExist f = MockFS $ do
    f <- gets ( mfs_get_contents f )
    return $ case f of
      Nothing -> False
      _ -> True
  
  fileSize f = processLengthAndBLFromFile f const
  
  getCurrentDirectory = return "/"

  getDirectoryContents dir = do
    dirs <- gets mfs_dirs
    files <- gets mfs_files
    return $ mapMaybe ( getRelative $ splitDirectories dir ) $ S.toList $ S.fromList $ dirs ++ files
    where
      getRelative pfx path =
        stripPrefix pfx ( splitDirectories path ) >>= singleToJust
      singleToJust [a] = Just a
      singleToJust _ = Nothing
  
  createDirectoryIfMissing True dir = modify $ mfs_mkdir dir
  createDirectoryIfMissing False dir =
    gets mfs_dirs >>= \ dirs -> if takeDirectory dir `elem` dirs
                             then modify $ mfs_mkdir dir
                             else error $ "Unable to create " ++ dir

  myCatch act _ = act

==> file4life-1.1.4-source.tgz: file4life-1.1.4/src/F4L/MockFSMonad.hs <==
{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}

module F4L.MockFSMonad where

import F4L.BufferMonad
import F4L.FSMonad
import F4L.Utils( uuidString )

import Control.DeepSeq
import Control.Lens
import Control.Monad.Error( throwError )
import Control.Monad.State.Strict
import qualified Data.ByteString.Lazy as BL
import Data.List( inits, stripPrefix )
import qualified Data.Map.Strict as M
import Data.Maybe( fromMaybe, mapMaybe )
import qualified Data.Set as S
import Data.Set( Set )
import System.Console.Haskeline.MonadException
import System.FilePath( takeDirectory, joinPath, splitDirectories )


subpaths :: FilePath -> [ FilePath ]
subpaths = map joinPath . drop 1 . inits . splitDirectories

{------------------------------ FS ------------------------------}

class MockROFSState mfs where
  mfs_files :: mfs -> [FilePath]
  mfs_get_contents :: FilePath -> mfs -> Maybe BL.ByteString
  mfs_dirs :: mfs -> [FilePath]
  mfs_is_deleted :: FilePath -> mfs -> Bool
  mfs_is_deleted _ _ = False


class MockROFSState mfs => MockFSState mfs where
  mfs_set_contents :: FilePath -> BL.ByteString -> mfs -> mfs
  mfs_delete :: FilePath -> mfs -> mfs
  mfs_mkdir :: FilePath -> mfs -> mfs
  mfs_now :: Lens' mfs Int


mfs_modifyContents
  :: MockFSState mfs   
  => FilePath -> ( BL.ByteString -> BL.ByteString ) -> mfs -> mfs
mfs_modifyContents p f fs = mfs_set_contents p ( f $ fromMaybe BL.empty $ mfs_get_contents p fs ) fs

mfs_mkdirAndParents :: MockFSState mfs => FilePath -> mfs -> mfs
mfs_mkdirAndParents f fs =   foldr mfs_mkdir fs $ subpaths f


{------------------------------ Overlay of 2 FSs ------------------------------}

-- | The first parameter overrides the second
data Overlay a b = Overlay a b

instance ( MockROFSState a, MockROFSState b ) => MockROFSState ( Overlay a b )
  where
    mfs_files ( Overlay a b ) = S.toList $ S.fromList $ mfs_files a ++ filter ( not . flip mfs_is_deleted a ) ( mfs_files b )
    mfs_get_contents f ( Overlay a b ) = mfs_get_contents f a `mplus` mfs_get_contents f b
    mfs_dirs ( Overlay a b ) = S.toList $ S.fromList $ mfs_dirs a ++ mfs_dirs b
    mfs_is_deleted f ( Overlay a _ ) = mfs_is_deleted f a


instance ( MockFSState a, MockROFSState b ) => MockFSState ( Overlay a b )
  where
    mfs_set_contents f s ( Overlay a b ) = Overlay ( mfs_set_contents f s a ) b
    mfs_delete f ( Overlay a b ) = Overlay ( mfs_delete f a ) b
    mfs_mkdir f ( Overlay a b ) = Overlay ( mfs_mkdir f a ) b
    mfs_now = lens f g . mfs_now
      where f ( Overlay a _ ) = a
            g ( Overlay _ b ) a = Overlay a b


{------------------------------ Virtual FS ------------------------------}

data VirtualFS = VirtualFS
                 FilePath -- root
                 Int -- number of files

instance MockROFSState VirtualFS where
  mfs_files ( VirtualFS root n ) = [ joinPath [ root, show i ] | i <- [ 1 .. n ] ]
  mfs_dirs ( VirtualFS root _ ) = subpaths root
  mfs_get_contents f ( VirtualFS root n ) =
    case splitAt ( length root ) f of
      (a, '/' : x @ (_ : _)) | a == root ->
        if all ( `elem` ['0' .. '9'] ) x && ( read x >= 1 ) && ( read x <= n )
           then Just $ BL.pack $ map ( toEnum . fromEnum ) $ "File " ++ show ( read x :: Int )
                else Nothing
      _ -> Nothing

{------------------------------ MemFS ------------------------------}

data MemFS = MemFS { _ms_files :: M.Map FilePath BL.ByteString
                   , _ms_deleted :: Set FilePath
                   , _ms_dirs :: Set FilePath
                   , _ms_now :: Int
                   }
           deriving Show

instance MockROFSState MemFS where
  mfs_files = M.keys . view ms_files
  mfs_get_contents f = view $ ms_files . to (M.lookup f)
  mfs_dirs = S.toList . view ms_dirs
  mfs_is_deleted f = view $ ms_deleted . to (S.member f)


instance MockFSState MemFS where
  mfs_set_contents f b =
    BL.length b `seq` over ms_files $ M.insert f b
  mfs_delete f = over ms_files ( M.delete f ) . over ms_deleted ( S.insert f )
  mfs_mkdir p = over ms_dirs $ S.insert p
  mfs_now = lens _ms_now $ \ s a -> s { _ms_now = a }


emptyMemFS = MemFS M.empty S.empty ( S.singleton "/" ) 0


ms_files :: Lens' MemFS ( M.Map FilePath BL.ByteString )
ms_files = lens _ms_files ( \ s a -> s { _ms_files = a } )
ms_dirs :: Lens' MemFS ( Set FilePath )
ms_dirs = lens _ms_dirs ( \ s a -> s { _ms_dirs = a } )
ms_deleted :: Lens' MemFS ( Set FilePath )
ms_deleted = lens _ms_deleted ( \ s a -> s { _ms_deleted = a } )


{------------------------------ Monad ------------------------------}

-- | Monad implementing FSMonad via an in-memory data structure
newtype MockFS fs m a = MockFS { deMockFS :: StateT fs ( IOWithBuffer m ) a }
                    deriving ( Monad, MonadIO, MonadState fs, BufferMonad, ParFSMonad )


runMockFS :: MonadIO m => MockFS fs m a -> StateT fs m a
runMockFS a = StateT $ runIOWithBuffer . runStateT ( deMockFS a )


instance MonadIO m => MonadException ( MockFS mfs m ) where
  controlIO _ = error "controlIO for MockFS not implemented"


instance ( MockFSState mfs, MonadIO m ) => DelFSMonad ( MockFS mfs m ) where
  removeFile = modify . mfs_delete


instance ( MockFSState mfs, MonadIO m ) => FSMonad ( MockFS mfs m ) where
  outputMsg = liftIO . outputMsg

  exifProperties = const $ return []

  prompt = liftIO . prompt

  generateUuid = liftIO uuidString

  getNow = use mfs_now

  readFromFile f = do
    mf <- gets $ mfs_get_contents f
    case mf of
      Nothing -> throwError $ "File " ++ f ++ " not found"
      Just s -> return $ map (toEnum . fromEnum) $ BL.unpack s

  actOnLengthAndBLFromFile f action = do
    mf <- gets $ mfs_get_contents f
    case mf of
      Nothing -> throwError $ "File " ++ f ++ " not found"
      Just s -> lift $ action ( fromIntegral $ BL.length s ) s
      
  appendToFile f s = do
    pre <- gets $ fromMaybe BL.empty . mfs_get_contents f
    modify $ mfs_set_contents f $ BL.append pre $ BL.pack $ map (toEnum . fromEnum) s
    -- re-read to make behaviour strictier. TODO: check if performance hit
  
  appendBLsToFile f bls = do
    pre <- gets $ fromMaybe BL.empty . mfs_get_contents f
    modify $ mfs_set_contents f $ foldl BL.append pre bls
  
  canonicalizePath = return -- liftIO . canonicalizePath
  
{-
  copyFile a b = do
    s <- gets $ mfs_get_contents a
    case s of
      Nothing -> fail $ "No such file " ++ a
      Just r -> modify $ mfs_set_contents b r
-}

  doesDirectoryExist d = gets mfs_dirs >>= return . elem d
  
  doesFileExist f = MockFS $ do
    f <- gets ( mfs_get_contents f )
    return $ case f of
      Nothing -> False
      _ -> True
  
  fileSize f = processLengthAndBLFromFile f const
  
  getCurrentDirectory = return "/"

  getDirectoryContents dir = do
    dirs <- gets mfs_dirs
    files <- gets mfs_files
    return $ mapMaybe ( getRelative $ splitDirectories dir ) $ S.toList $ S.fromList $ dirs ++ files
    where
      getRelative pfx path =
        stripPrefix pfx ( splitDirectories path ) >>= singleToJust
      singleToJust [a] = Just a
      singleToJust _ = Nothing
  
  createDirectoryIfMissing True dir = modify $ mfs_mkdir dir
  createDirectoryIfMissing False dir =
    gets mfs_dirs >>= \ dirs -> if takeDirectory dir `elem` dirs
                             then modify $ mfs_mkdir dir
                             else error $ "Unable to create " ++ dir

  myCatch act _ = act
